
########################################################################################
###
###                         "MFPCA-sparse-new.R" 
###
###         This file contains R function "MFPCA.sparse" to implement Multilevel Functional Principal Component Analysis
###         for sparse multilevel functional data. This is the new version, and can handle large data well. 
###                                     
###         Author:     Chongzhi Di
###         Date:       08/16/2009
###
###                                                             
###         References:
###             (1) Di, Crainiceanu, Caffo and Punjabi (2009), Multilevel Functional Principal Component Analysis.
###                 The Annals of Applied Statistics, Vol. 3, No. 1, 458488.
###             (2) Crainiceanu, Staicu and Di (2009), Generalized Multilevel Functional Regression. JASA, to appear.
###             (3) Di and Crainiceanu (2009), Multilevel Functional Principal Component Analysis for Sparsely Sampled
###                 Hierarchical Functions. In preparation for submission. 
###            
########################################################################################






#########################################################################################
###
###     The function "unique.average" find unique categories according to "index" variable, 
###                         and take averages in variable "y" for cells with the same index.
###
###         For example:
###             index = c(1, 1, 2, 3, 3, 3, 5)
###             y     = c(0, 2, 7, 4, 8, 6, 9)
###         Applying "unique.average" to this data
###             fit <- unique.average(index, y)
###         The results would be: 
###             fit$index = c(1, 2, 3, 5)
###             fit$y     = c(1, 7, 6, 9)
###
#########################################################################################

unique.average <- function(index, y) {
    
    n <- sum(!is.na(y))
    y.new <- rep(NA, n)
    index.new <- rep(NA, n)
    count <- rep(0, n)
    
     
    y.new[1] <- y[1]
    index.new[1] <- index[1] 
    count[1] <- 1
    current <- 1
    
    if(n > 1) {
        for(i in 2:n) {
            if(index[i]==index[i-1]) {
                count[i] <- count[i-1] + 1
                y.new[current] <- ( y.new[current] * count[i-1] + y[i] )/count[i]
                count[i-1] <- count[i]
                   
            }
            else {
                current <- current + 1
                y.new[current] <- y[i] 
                count[i] <- 1
                index.new[current] <- index[i] 
            }
        }
    }
    
    return(list(index=index.new, y=y.new))
}




##################################################################################################
### 
###     "MFPCA.sparse" Verson 2.0
###
###     Description:
###         The function "MFPCA.sparse" is designed to implement multilevel functional principal 
###         component analysis for sparse data. The methodology aim to separate the variation in 
###         multilevel functional data into two parts, between cluster and within cluster variations. 
###         The principal directions of variation at each level is characterized by a set of 
###         eigenfunctions, ordered by the eigenvalues that indicate the magnate of variation.
###
###     Improvements over Version 1.0:
###         This version can handle large data size reasonably well. The main problem with Version 1.0
###         is the large memory needed in bivariate smoothing for covariance operators. In this version,
###         we first divide the time interval into 100 bins, take averages over each bin, and then
###         conduct bivariate smoothing over 10000=100*100 cells. This would control the computational 
###         burden, even if actual data have larger size. 
###
##################################################################################################


########################################################################################
###
### Input of the function "MFPCA.sparse"
###
### y:      a data matrix that contains functional responses. Each row contains measurements 
###         from a function at a set of grid points, and each column contains measurements
###         of all functions at a particular grid point.
### tx:     a data matrix that specifies grid points for each functions. 
### id:     a vector contains the id information that is used to identify clusters. 
### visit:  optional argument if functions from the same cluster have natural orders 
###         such as visit time.
### n.grid: a vector, each element is the number of grid points for a function (one row of y)
### N:      the maximum number of grid points, i.e, the number of columns of y and tx.
###
### tstart: the start point of the grid points
### tlength: the length of the interval that the functions are evaluated at.
### twoway: a logical argument indicates whether oneway or twoway functional ANOVA (analysis
###         of variance) decomposition is more appropriate for the problem. "twoway=TRUE" will carry
###         out twoway ANOVA, and the default, "twoway=FALSE" will carry out oneway ANOVA.
### min.K1: the minimum number of components to keep in level 1 space. The actual number of 
###         components will be the maximum of min.K1 and the number determined by the algorithm.
### min.K2: the minimum number of components to keep in level 2 space.
###
### N.smooth.single:    the number of knots for univariate smoothing
### N.smooth:           the number of knots per dimension for bivariate smoothing
### N.fit:              the number of grid points over which the mean functions and eigenfunctions are estimated
### 
########################################################################################


########################################################################################
###
### Output of the function "MFPCA.sparse"
### 
###     The output of the function is a list that contains the following elements. 
###
###
### t.fit:      the set of grid points over which the mean functions and eigenfunctions are estimated
### N.fit:      the number of grid points in "t.fit" (this is the same as the input "N.fit")
###
### K1:         the number of components at level 1
### K2:         the number of components at level 2
### lambda1:    a vector that includes all level 1 eigenvalues in non-increasing order
### lambda2:    a vector that includes all level 2 eigenvalues in non-increasing order
### phi1:       a matrix that contains all level 1 eigenfunctions. Each row corresponds an 
###             eigenfunction evaluated at the same set of grid points as the input data.
###             The eigenfunctions are in the same order as the corresponding eigenvalues.
### phi2:       a matrix that contains all level 2 eigenfunctions. Each row corresponds an 
###             eigenfunction evaluated at the same set of grid points as the input data.
###             The eigenfunctions are in the same order as the corresponding eigenvalues.
###
### mu:         the estiamted overall mean function, evaluated at the grid points "t.fit"
### eta:        the estimated visit specific shifts from overall mean, evaluated at the grid points "t.fit"
###             an "N.fit" times "J" matrix, where each column corresponds to a visit and each row 
###                                         corresponds to a grid point. 
### var.noise:  the variance of the residual measurement error 
### y.hat:      estimated curves for all subjects and visits, evaluated at the grid points "t.fit"
###             an "M*J" times "N.fit" matrix, with similar structure to the input functions "y"
### y.hat.sd:   estimated pointwise standard errors for "y.hat"
### y.hat.subject:  estimated subject specific curves for all subjects, evaluated at the grid points "t.fit"
###                 an "M" times "N.fit" matrix, each row of which corresponds to a subject
### y.hat.subject.sd: estimated pointwise standard errors for "y.hat.subject"

###
#########################################################################################


MFPCA.sparse <- function(y, tx, id=NULL, visit=NULL, M=NULL, J=2, n.grid=NULL, N=NULL, N.fit=50, tstart=0, tlength=1, 
                             min.K1=4, min.K2=4, N.smooth.single=20, N.smooth=15, twoway=FALSE) {

###     check the missingness of arguments
###     1. Either (id,visit) or (M,J) is needed for identifying clusters. 
###         If both are missing, give error message "not enough information";
###         if only (M,J) is provided, the function would generate (id,visit);
###         if only ((id,visit) is provided, the function would generate (M,J);
###      Note:  currently the function only deals with balanced data, i.e, each 
###             subject is measured at the same J visits.
###     2. If the number of grid points N is missing, the function would generate it.

    if(  ( is.null(id)|is.null(visit) ) && ( is.null(M)|is.null(J) )  )
        stop("Not enough information! Please provide (id, visit) or (M,J) !")
    if(is.null(n.grid))
        n.grid <- apply(!is.na(y), 1, sum)
    if(is.null(N)) 
        N <- max(n.grid.obs)
        
    n.visit <- rep(J, M)
        

###     Generate the set of grid points that are equally spaced, on which we would
###     predit the mean and covariance functions.
    t.knot <- seq(tstart, tstart + tlength, length=N.smooth.single)
    t.fit <- seq(tstart, tstart + tlength, length=N.fit)
    
    N.row <- dim(y)[1]
    
    y1 <- matrix(0, M*J, N)
    index <- matrix(0, M*J, N)
    t.fit.matrix <- matrix(rep(t.fit, each=N), nrow=N)

    for(m in 1:M ) {
        for(j in 1:J) {
            row.ind <- (m-1)*J + j
            index.temp <- apply( (t.fit.matrix - tx[row.ind, ])^2, 1, which.min )
            fit <- unique.average(index.temp, y[row.ind,] )
            y1[row.ind,] <- fit$y
            index[row.ind, ] <- fit$index
        }
    }
    
    n.grid <- apply(!is.na(y1), 1, sum)

    
###     Calculate overall mean function
    library(SemiPar)
    mu.fit <- rep(0,  N.fit)
    eta.fit <- matrix(0, N.fit, J)
    mueta.fit <- matrix(0, N.fit, J)
    
    mu.obs <- matrix(0, M*J, N)
    mueta.obs <- matrix(0, M*J, N)
    
    ###  divide the interval into "N.fit" bins to reduce computational burden for smoothing 
    mu.sum <- rep(0, N.fit)
    mu.count <- rep(0, N.fit) 
    mu.mean <- rep(0, N.fit)
    for (m in 1:M) {
        for(j in 1:J) {
            row.ind <- (m-1)*J + j
            row.len <- n.grid[row.ind]
            mu.count[ index[row.ind, 1:row.len] ] <- mu.count[ index[row.ind, 1:row.len] ] + 1
            mu.sum[ index[row.ind, 1:row.len] ] <- mu.sum[ index[row.ind, 1:row.len] ] + y1[row.ind, 1:row.len ]
        }
    } 
    mu.mean <- ifelse(mu.count==0, NA,  mu.sum/mu.count)    
    
        temp.data <- data.frame(y2 = mu.mean,  x2 = t.fit )
        attach(temp.data, pos=3)
        fit.mu <- spm( y2~f(x2,knots=t.knot) ,omit.missing=TRUE)
        newx <- data.frame(x2=t.fit)
        mu.fit <- predict(fit.mu,newx)
        detach(pos=3)
        mu.obs <- matrix( mu.fit[ index ], ncol=N)
    
    if(twoway==TRUE) {
        for(j in 1:J) {
            mu.sum <- rep(0, N.fit)
            mu.count <- rep(0, N.fit) 
            mu.mean <- rep(0, N.fit)
            for (m in 1:M) {
                row.ind <- (m-1)*J + j
                row.len <- n.grid[row.ind]
                mu.count[ index[row.ind, 1:row.len] ] <- mu.count[ index[row.ind, 1:row.len] ] + 1
                mu.sum[ index[row.ind, 1:row.len] ] <- mu.sum[ index[row.ind, 1:row.len] ] + y1[row.ind, 1:row.len ]
            } 
            mu.mean <- ifelse(mu.count==0, NA,  mu.sum/mu.count)           
                        
            temp.data2 <- data.frame(y3 = mu.mean, x3 = t.fit )
            attach(temp.data2, pos=4)
            fit.eta <- spm( y3~f(x3,knots=t.knot) ,omit.missing=TRUE)
            newx <- data.frame(x3=t.fit)
            mueta.fit[,j] <- predict(fit.eta,newx)
            eta.fit[,j] <- mueta.fit[,j] - mu.fit
            detach(pos=4)
            ind.visit.j <-  (0:(M-1)*J) + j
            mueta.obs[ ind.visit.j, ] <- matrix( mueta.fit[ index[ind.visit.j,], j], ncol=N)
        }
    }
    
    ### test only
        #plot(t.fit, mu.fit, col="red", lwd=3, type="l") 
        #lines(t.fit, mueta.fit[,1], col="red")
        #lines(t.fit, mueta.fit[,2], col="red")
        #lines(t.fit, mu.true(t.fit), col="blue", lwd=2)
    ### end test only


###    Calculate residuals by subtracting visit-specific mean from original functions for 
###    "twoway=TRUE", or subtracting overall mean function for "twoway=FALSE". 
    resd <- matrix(0, nrow=M*J, ncol=N) 
    resd <- y - mu.obs
    if(twoway==TRUE) {
        resd <- y1 - mueta.obs
    }





###     Calculate the pairs to calculate the total covariance function
###         the diagonal terms are already excluded

    ###  divide the interval into "N.fit" bins to reduce computational burden for smoothing 
    cov.sum <- matrix(0, N.fit, N.fit)
    cov.count <- matrix(0, N.fit, N.fit) 
    cov.mean <-  matrix(0, N.fit, N.fit)
    for (m in 1:M) {
        for(j in 1:J) {
            row.ind <- (m-1)*J + j
            len.ind <- n.grid[row.ind]
            temp.ind <- index[row.ind, 1:len.ind]
            cov.count[ temp.ind, temp.ind ] <- cov.count[ temp.ind, temp.ind ] + 1
            cov.sum[ temp.ind, temp.ind ] <- cov.sum[ temp.ind, temp.ind ] + resd[row.ind, 1:len.ind ] %*% t( resd[row.ind, 1:len.ind])
        }
    } 
    cov.mean <- ifelse(cov.count==0, NA,  cov.sum/cov.count)    

    cov.mean.nodiag <- cov.mean
    diag(cov.mean.nodiag) <- rep(NA, N.fit)
    resd.pair <- cbind( as.vector(cov.mean.nodiag), rep(t.fit, each=N.fit), rep(t.fit, N.fit) )
    resd.pair.diag <- cbind( diag(cov.mean), t.fit) 

    
###     Calculate the pairs to calculate the between covariance funcion
###     first, calculate the possible pairs of same-subject visits

    cov.sum <- matrix(0, N.fit, N.fit)
    cov.count <- matrix(0, N.fit, N.fit) 
    cov.mean <-  matrix(0, N.fit, N.fit)

    for(m in 1:M) {
        for(j in 1:(n.visit[m]-1) ) 
            row.ind1 <- (m-1)*J + j                
            len.ind1 <- sum( ! is.na( index[row.ind1,] ) )
            temp.ind1 <- index[row.ind1, 1:len.ind1]
            for(k in (j+1):n.visit[m] ) {
                row.ind2 <- (m-1)*J + k
                len.ind2 <- sum( ! is.na( index[row.ind2,] ) )
                temp.ind2 <- index[row.ind2, 1:len.ind2]
                cov.count[ temp.ind1, temp.ind2 ] <- cov.count[ temp.ind1, temp.ind2 ] + 1
                cov.sum[ temp.ind1, temp.ind2 ] <- cov.sum[ temp.ind1, temp.ind2 ] + resd[row.ind1, 1:len.ind1 ] %*% t(resd[row.ind2, 1:len.ind2] )  
                cov.count[ temp.ind2, temp.ind1 ] <- cov.count[ temp.ind2, temp.ind1 ] + 1
                cov.sum[ temp.ind2, temp.ind1 ] <- cov.sum[ temp.ind2, temp.ind1 ] + resd[row.ind2, 1:len.ind2 ] %*% t(resd[row.ind1, 1:len.ind1] )                                                        
            }
    }
    cov.mean <- ifelse(cov.count==0, NA,  cov.sum/cov.count)    
    resd.pair.between <- cbind( as.vector(cov.mean), rep(t.fit, each=N.fit), rep(t.fit, N.fit) )


###     Estimate the three covariance functions: overall covariance G, 
###     between covariance Gb and within covariance Gw
    G <- matrix(0, N.fit, N.fit)
    Gb <- matrix(0, N.fit, N.fit)
    Gw <- matrix(0, N.fit, N.fit)
        
    ### load the "SemiPar" package which contains functions for semiparametric smoothing 
    ### using linear mixed models
        t.smooth <- seq(tstart, tstart + tlength ,length=N.smooth)
        
        myknots.t <- data.frame(x1t=rep(t.smooth, each=N.smooth), x2t=rep(t.smooth, N.smooth))         
        data.gt <- data.frame(gt=resd.pair[,1], x1t=resd.pair[,2], x2t=resd.pair[,3])
        attach(data.gt, pos=4)                  
        fit.t <- spm(gt ~ f(x1t, x2t, knots=myknots.t),omit.missing=TRUE)
        newdata.t <- data.frame(x1t=rep(t.fit, each=N.fit),x2t=rep(t.fit, N.fit))
        pred.t <- predict(fit.t, newdata.t)
        detach(pos=4)

        myknots.b <- data.frame(x1b=rep(t.smooth, each=N.smooth), x2b=rep(t.smooth, N.smooth))
        data.gb <- data.frame(gb=resd.pair.between[,1], x1b=resd.pair.between[,2], x2b=resd.pair.between[,3])
        attach(data.gb, pos=5) 
        fit.b <- spm(gb ~ f(x1b, x2b,knots=myknots.b),omit.missing=TRUE)        
        newdata.b <- data.frame(x1b=rep(t.fit, each=N.fit),x2b=rep(t.fit, N.fit))
        pred.b <- predict(fit.b, newdata.b)
        detach(pos=5)
        
        s.gb <- matrix(pred.b, N.fit)
        Gb <- ( s.gb + t(s.gb) )/2
        s.gw <- matrix(pred.t - pred.b, N.fit)
        Gw <- ( s.gw + t(s.gw) )/2
        
        
    ### calculate the noise variance 
        temp.data <- data.frame(var.y = as.vector(resd^2),  var.x = as.vector(tx) )
        attach(temp.data, pos=6)
        fit.var <- spm( var.y ~ f(var.x,knots=t.knot) ,omit.missing=TRUE)
        newx <- data.frame(var.x=t.fit)
        var.fit <- predict(fit.var,newx)          
    
        var.noise <- mean( var.fit - diag(matrix(pred.t,N.fit)) ) 
    



###     Estimate eigen values and eigen functions at two levels by calling the 
###     eigen function (in R "base" package) on discretized covariance matrices.
    e1 <- eigen(Gb)
    e2 <- eigen(Gw)

###    transform the eigenvalues of the covariance matrices into eigenvalues
###    of original functions
    fpca1.value <- e1$values* tlength / N.fit
    fpca2.value <- e2$values* tlength / N.fit 
###     Keep only non-negative eigenvalues
    fpca1.value <- ifelse(fpca1.value>=0, fpca1.value, 0)
    fpca2.value <- ifelse(fpca2.value>=0, fpca2.value, 0)
###     Calculate the percentage of variance that are explained by the components
    percent1 <- (fpca1.value)/sum(fpca1.value)
    percent2 <- (fpca2.value)/sum(fpca2.value)
    
###     Decide the number of components that are kept at level 1 and 2. The general
###     rule is to stop at the component where the cumulative percentage of variance 
###     explained is greater than 90% and the variance explained by any single component
###     after is less than 1/N. The number of components are also no less than the 
###     pre-determined minimum values min.K1 or min.K2.
    K1 <- max( which(cumsum(percent1) < 0.9 | percent1 > 1/N.fit ) + 1, min.K1 )
    K2 <- max( which(cumsum(percent2) < 0.9 | percent2 > 1/N.fit ) + 1, min.K2 )

###     estimate eigen vectors for discretized covariance matrices and
###     transform them into norm one eigenfunctions
    fpca1.vectors <- e1$vectors[, 1:K1]*sqrt(N.fit/tlength)
    fpca2.vectors <- e2$vectors[, 1:K2]*sqrt(N.fit/tlength)


###     The eigenfunctions are unique only up to a change of signs.
###     Select the signs of eigenfunctions so that the integration over the domain 
###     is non-negative
    for(i in 1:K1) {
        v2 <- fpca1.vectors[,i]
        tempsign <- sum(v2)
        fpca1.vectors[,i] <- ifelse(tempsign<0, -1,1) * v2
    }
    for(i in 1:K2) {
        v2 <- fpca2.vectors[,i]
        tempsign <- sum(v2)
        fpca2.vectors[,i] <- ifelse(tempsign<0, -1,1) * v2
    }


lam1=fpca1.value[1:K1]
lam2=fpca2.value[1:K2]
phi1=fpca1.vectors
phi2=fpca2.vectors


###
###     Estimatimate principal component scores and predict various functions
###

library(MASS) 

index <- matrix(0, M*J, N)
mu.hat <- matrix(0, M*J, N)    
y.center <- matrix(0, M*J, N)

phi1.hat <- array(0, dim=c( M*J, N, K1) )
phi2.hat <- array(0, dim=c( M*J, N, K2) )

t.fit.matrix <- matrix(rep(t.fit, each=N), nrow=N)

for(m in 1:M ) {
    for(j in 1:J) {
        row.ind <- (m-1)*J + j
        index <- apply( (t.fit.matrix - tx[row.ind, ])^2, 1, which.min )
        y.center[row.ind, ] <- y[ row.ind ,] - mu.fit[ index ] - eta.fit[ index , j] 
        phi1.hat[row.ind, , ] <- phi1[index, ]
        phi2.hat[row.ind, , ] <- phi2[index, ]
    }
}



est.si1 <- matrix(0, nrow=M, ncol=K1)
est.si2 <- matrix(0, nrow=M*J, ncol=K2)

est.si1.sd <- matrix(0, nrow=M, ncol=K1)
est.si2.sd <- matrix(0, nrow=M*J, ncol=K2)


y.hat.center <- matrix(0, M*J, N.fit) 
y.hat <- matrix(0, M*J, N.fit)
y.hat.subject <- matrix(0, M, N.fit)

y.hat.sd <- matrix(0, M*J, N.fit)
y.hat.subject.sd <- matrix(0, M, N.fit)



for(m in 1:M) {

cov.y <- matrix(0, N*J, N*J)
cov.xi.y <- matrix(0, K1 + J*K2,  N*J )

    for(j1 in 1:J) {
        ind1 <- J*(m-1) + j1
        cov.y [ 1:N + (j1-1) *N , 1:N + (j1-1) *N ] <- phi1.hat[ind1,,] %*% diag( lam1) %*%  t( phi1.hat[ind1,,] ) + phi2.hat[ind1,,] %*% diag( lam2) %*%  t( phi2.hat[ind1,,] ) + diag( rep(var.noise, N) )
        
        if(j1<J) {
            for(j2 in (j1+1):J ) {
                ind2 <- J*(m-1) + j2
                cov.y[ 1:N + (j2-1) *N , 1:N + (j1-1) *N ] <- phi1.hat[ind2,,] %*% diag( lam1) %*%  t( phi1.hat[ind1,,] )
                cov.y[ 1:N + (j1-1) *N , 1:N + (j2-1) *N ] <- t( cov.y[ 1:N + (j2-1) *N , 1:N + (j1-1) *N ] )
            }
        } 
    }

    for(j in 1:J) {
        ind1 <- J*(m-1) + j
        cov.xi.y[1:K1, 1:N + (j-1) *N] <- diag( lam1) %*% t( phi1.hat[ind1,,] )
        cov.xi.y[ K1 + (1:K2) + K2* (j-1), 1:N + N*(j-1) ] <- diag( lam2) %*% t( phi2.hat[ind1,,] )  
    }

temp.mat <- cov.xi.y %*% ginv(cov.y)
score <- temp.mat %*% as.vector( t(y.center[J*(m-1) + 1:J,]) )
var.score <- diag( c(lam1, rep(lam2, J) ) ) -  temp.mat %*% t(cov.xi.y)

est.si1[m ,] <- score[1:K1]
est.si2[ 2*(m-1) + 1:J  ,] <- t( matrix( score[ K1 + (1: (K2*J) ) ], ncol=J) )

y.hat.subject[m, ] <- mu.fit + eta.fit[,j] + phi1 %*% est.si1[m,] 
y.hat.subject.sd[m,] <-  sqrt( diag( phi1 %*% var.score[1:K1, 1:K1] %*% t( phi1 ) ) )

    for(j in 1:J) {
        row.ind <- (m-1)*J + j
        y.hat.center[ row.ind ,] <-  phi1 %*% est.si1[m,] + phi2 %*% est.si2[row.ind ,] 
        y.hat[row.ind,] <- y.hat.center[ row.ind ,]  +  mu.fit + eta.fit[,j] 
        
        temp1 <- cbind( phi1, phi2 )
        temp2 <- var.score[ c(1:K1, K1 + (j-1)* K2 + (1:K2)  ) , c(1:K1, K1 + (j-1)* K2 + (1:K2)  ) ]
        y.hat.sd[row.ind, ] <- sqrt( diag( temp1 %*% temp2 %*% t(temp1) ) )   
    }

}


###     Return the results from multilevel FPCA as a list
    return( list(K1=K1, K2=K2, t.fit=t.fit, N.fit=N.fit, lambda1=fpca1.value, lambda2=fpca2.value, phi1=fpca1.vectors, 
        phi2=fpca2.vectors, mu=mu.fit, eta=eta.fit, var.noise=var.noise, y.hat=y.hat, y.hat.sd=y.hat.sd, 
        y.hat.subject=y.hat.subject, y.hat.subject.sd=y.hat.subject.sd ) )

} 
