
########################################################################################
###
###                         "MFPCA-sparse.R" 
###
###         This file contains R function "MFPCA.sparse" to implement Multilevel Functional Principal Component Analysis
###         for sparse multilevel functional data. This is the old version, and may need a lot of memory space for 
###         medium to large data. 
###                                     
###         Author:     Chongzhi Di
###         Date:       08/10/2009
###
###                                                             
###         References:
###             (1) Di, Crainiceanu, Caffo and Punjabi (2009), Multilevel Functional Principal Component Analysis.
###                 The Annals of Applied Statistics, Vol. 3, No. 1, 458488.
###             (2) Crainiceanu, Staicu and Di (2009), Generalized Multilevel Functional Regression. JASA, to appear.
###             (3) Di and Crainiceanu (2009), Multilevel Functional Principal Component Analysis for Sparsely Sampled
###                 Hierarchical Functions. In preparation for submission. 
###            
########################################################################################


########################################################################################
### 
###     "MFPCA.sparse" Verson 1.0
###
###     Description:
###         The function "MFPCA.sparse" is designed to implement multilevel functional principal 
###         component analysis for sparse data. The methodology aim to separate the variation in 
###         multilevel functional data into two parts, between cluster and within cluster variations. 
###         The principal directions of variation at each level is characterized by a set of 
###         eigenfunctions, ordered by the eigenvalues that indicate the magnate of variation.
###
###     Note: 
###         This version may need a log ot memory space for medium to large size data. 
###         For large data, we suggest the use of Version 2 instead. 
###
########################################################################################


########################################################################################
###
### Input of the function "MFPCA.sparse"
###
### y:      a data matrix that contains functional responses. Each row contains measurements 
###         from a function at a set of grid points, and each column contains measurements
###         of all functions at a particular grid point.
### tx:     a data matrix that specifies grid points for each functions. 
### id:     a vector contains the id information that is used to identify clusters. 
### visit:  optional argument if functions from the same cluster have natural orders 
###         such as visit time.
### n.grid: a vector, each element is the number of grid points for a function (one row of y)
### N:      the maximum number of grid points, i.e, the number of columns of y and tx.
###
### tstart: the start point of the grid points
### tlength: the length of the interval that the functions are evaluated at.
### twoway: a logical argument indicates whether oneway or twoway functional ANOVA (analysis
###         of variance) decomposition is more appropriate for the problem. "twoway=TRUE" will carry
###         out twoway ANOVA, and the default, "twoway=FALSE" will carry out oneway ANOVA.
### min.K1: the minimum number of components to keep in level 1 space. The actual number of 
###         components will be the maximum of min.K1 and the number determined by the algorithm.
### min.K2: the minimum number of components to keep in level 2 space.
###
### N.smooth.single:    the number of knots for univariate smoothing
### N.smooth:           the number of knots per dimension for bivariate smoothing
### N.fit:              the number of grid points over which the mean functions and eigenfunctions are estimated
### 
########################################################################################


########################################################################################
###
### Output of the function "MFPCA.sparse"
### 
###     The output of the function is a list that contains the following elements. 
###
###
### t.fit:      the set of grid points over which the mean functions and eigenfunctions are estimated
### N.fit:      the number of grid points in "t.fit" (this is the same as the input "N.fit")
###
### K1:         the number of components at level 1
### K2:         the number of components at level 2
### lambda1:    a vector that includes all level 1 eigenvalues in non-increasing order
### lambda2:    a vector that includes all level 2 eigenvalues in non-increasing order
### phi1:       a matrix that contains all level 1 eigenfunctions. Each row corresponds an 
###             eigenfunction evaluated at the same set of grid points as the input data.
###             The eigenfunctions are in the same order as the corresponding eigenvalues.
### phi2:       a matrix that contains all level 2 eigenfunctions. Each row corresponds an 
###             eigenfunction evaluated at the same set of grid points as the input data.
###             The eigenfunctions are in the same order as the corresponding eigenvalues.
###
### mu:         the estiamted overall mean function, evaluated at the grid points "t.fit"
### eta:        the estimated visit specific shifts from overall mean, evaluated at the grid points "t.fit"
###             an "N.fit" times "J" matrix, where each column corresponds to a visit and each row 
###                                         corresponds to a grid point. 
### var.noise:  the variance of the residual measurement error 
### y.hat:      estimated curves for all subjects and visits, evaluated at the grid points "t.fit"
###             an "M*J" times "N.fit" matrix, with similar structure to the input functions "y"
### y.hat.sd:   estimated pointwise standard errors for "y.hat"
### y.hat.subject:  estimated subject specific curves for all subjects, evaluated at the grid points "t.fit"
###                 an "M" times "N.fit" matrix, each row of which corresponds to a subject
### y.hat.subject.sd: estimated pointwise standard errors for "y.hat.subject"

###
#########################################################################################


MFPCA.sparse <- function(y, tx, id=NULL, visit=NULL, M=NULL, J=2, n.grid=NULL, N=NULL, N.fit=50, tstart=0, tlength=1, 
                             min.K1=4, min.K2=4, N.smooth.single=20, N.smooth=15, twoway=FALSE) {

###     check the missingness of arguments
###     1. Either (id,visit) or (M,J) is needed for identifying clusters. 
###         If both are missing, give error message "not enough information";
###         if only (M,J) is provided, the function would generate (id,visit);
###         if only ((id,visit) is provided, the function would generate (M,J);
###      Note:  currently the function only deals with balanced data, i.e, each 
###             subject is measured at the same J visits.
###     2. If the number of grid points N is missing, the function would generate it.

    if(  ( is.null(id)|is.null(visit) ) && ( is.null(M)|is.null(J) )  )
        stop("Not enough information! Please provide (id, visit) or (M,J) !")
    if(is.null(n.grid))
        n.grid <- apply(!is.na(y), 1, sum)
    if(is.null(N)) 
        N <- max(n.grid.obs)
        
    n.visit <- rep(J, M)
        

###     Generate the set of grid points that are equally spaced, on which we would
###     predit the mean and covariance functions.
    t.knot <- seq(tstart, tstart + tlength, length=N.smooth.single)
    t.fit <- seq(tstart, tstart + tlength, length=N.fit)
    
###     Calculate overall mean function
    library(SemiPar)
    mu.fit <- rep(0,  N.fit)
    eta.fit <- matrix(0, N.fit, J)
    
    mu.obs <- matrix(0, M*J, N)
    mueta.obs <- matrix(0, M*J, N)
    
        temp.data <- data.frame(y2 =as.vector(y),  x2 = as.vector(tx) )
        attach(temp.data, pos=3)
        fit.mu <- spm( y2~f(x2,knots=t.knot) ,omit.missing=TRUE)
        newx <- data.frame(x2=t.fit)
        mu.fit <- predict(fit.mu,newx)
        detach(pos=3)
        mu.obs <- matrix(fit.mu$fit$fitted, nrow=M*J, ncol=N)
    
    if(twoway==TRUE) {
        for(j in 1:J) {
            temp.data2 <- data.frame(y3 = as.vector( y[ (0:(M-1)*J) + j,  ] ), x3 = as.vector( tx[ (0:(M-1)*J) + j,  ] ) )
            attach(temp.data2, pos=4)
            fit.eta <- spm( y3~f(x3,knots=t.knot) ,omit.missing=TRUE)
            newx <- data.frame(x3=t.fit)
            eta.fit[,j] <- predict(fit.eta,newx) - mu.fit
            detach(pos=4)
            mueta.obs[(0:(M-1)*J) + j, ] <- matrix(fit.eta$fit$fitted, nrow=M, ncol=N)
        }
    }


###    Calculate residuals by subtracting visit-specific mean from original functions for 
###    "twoway=TRUE", or subtracting overall mean function for "twoway=FALSE". 
    resd <- matrix(0, nrow=M*J, ncol=N) 
    resd <- y - mu.obs
    if(twoway==TRUE) {
        resd <- y - mueta.obs
    }


###     Calculate the pairs to calculate the total covariance function
###         the diagonal terms are already excluded
    resd.pair <- matrix(0, nrow=sum(n.grid*(n.grid-1)), ncol=3)
    resd.pair.diag <- matrix(0, nrow=sum(n.grid), ncol=2)
    cur.index <- 1
    cur.index.diag <- 1
    for(i in 1:(M*J) ) {
        len.pair <- n.grid[i]*(n.grid[i]-1)/2
        resd.pair.diag[cur.index.diag:(cur.index.diag+n.grid[i]-1),] <- cbind(resd[i,]^2, tx[i,])
        cur.index.diag <- cur.index.diag + n.grid[i] 
        for(j in 1:(n.grid[i]-1) ) 
            for(k in (j+1):n.grid[i]) {
                resd.pair[cur.index,] <-  c(resd[i,j]*resd[i,k], tx[i,j], tx[i,k])
                resd.pair[cur.index+1,] <- c(resd[i,j]*resd[i,k], tx[i,k], tx[i,j])
                cur.index <- cur.index+2
            }
    }
    
###     Calculate the pairs to calculate the between covariance funcion
###     first, calculate the possible pairs of same-subject visits
    index.pair.between <- matrix(0, nrow=sum(n.visit*(n.visit-1)/2),ncol=4) 
    cur.index <- 1
    sub.before <- 0
    for(i in 1:M) {
        for(j in 1:(n.visit[i]-1) ) 
            for(k in (j+1):n.visit[i] ) {
                index.pair.between[cur.index,] <- c(sub.before+j, sub.before+k, n.grid[sub.before+j], n.grid[sub.before+k])
                cur.index <- cur.index + 1                
            }
        sub.before <- sub.before + n.visit[i]
    }
    
    resd.pair.between <- matrix(0, nrow=sum(index.pair.between[,3]*index.pair.between[,4]), ncol=3)
    cur.index <- 0
    for(i in 1:M) {
        ni.pair <- index.pair.between[i,3]*index.pair.between[i,4]
        row.ind1 <- index.pair.between[i,1]
        row.ind2 <- index.pair.between[i,2]
        col.ind1 <- rep(1:index.pair.between[i,3], each=index.pair.between[i,4])
        col.ind2 <- rep(1:index.pair.between[i,4], index.pair.between[i,3])
        resd.pair.between[(cur.index+1): (cur.index+ni.pair), ] <- cbind(resd[row.ind1,col.ind1]*resd[row.ind2,col.ind2], tx[row.ind1,col.ind1], tx[row.ind2,col.ind2]   )
        cur.index <- cur.index + ni.pair
    }

###     Estimate the three covariance functions: overall covariance G, 
###     between covariance Gb and within covariance Gw
    G <- matrix(0, N.fit, N.fit)
    Gb <- matrix(0, N.fit, N.fit)
    Gw <- matrix(0, N.fit, N.fit)
        
    ### load the "SemiPar" package which contains functions for semiparametric smoothing 
    ### using linear mixed models
        t.smooth <- seq(tstart, tstart + tlength ,length=N.smooth)
        
        myknots.t <- data.frame(x1t=rep(t.smooth, each=N.smooth), x2t=rep(t.smooth, N.smooth))         
        data.gt <- data.frame(gt=resd.pair[,1], x1t=resd.pair[,2], x2t=resd.pair[,3])
        attach(data.gt, pos=4)                  
        fit.t <- spm(gt ~ f(x1t, x2t, knots=myknots.t),omit.missing=TRUE)
        newdata.t <- data.frame(x1t=rep(t.fit, each=N.fit),x2t=rep(t.fit, N.fit))
        pred.t <- predict(fit.t, newdata.t)
        detach(pos=4)

        myknots.b <- data.frame(x1b=rep(t.smooth, each=N.smooth), x2b=rep(t.smooth, N.smooth))
        data.gb <- data.frame(gb=resd.pair.between[,1], x1b=resd.pair.between[,2], x2b=resd.pair.between[,3])
        attach(data.gb, pos=5) 
        fit.b <- spm(gb ~ f(x1b, x2b,knots=myknots.b),omit.missing=TRUE)        
        newdata.b <- data.frame(x1b=rep(t.fit, each=N.fit),x2b=rep(t.fit, N.fit))
        pred.b <- predict(fit.b, newdata.b)
        detach(pos=5)
        
        s.gb <- matrix(pred.b, N.fit)
        Gb <- ( s.gb + t(s.gb) )/2
        s.gw <- matrix(pred.t - pred.b, N.fit)
        Gw <- ( s.gw + t(s.gw) )/2
        
        
    ### calculate the noise variance 
        temp.data <- data.frame(var.y = as.vector(resd^2),  var.x = as.vector(tx) )
        attach(temp.data, pos=6)
        fit.var <- spm( var.y ~ f(var.x,knots=t.knot) ,omit.missing=TRUE)
        newx <- data.frame(var.x=t.fit)
        var.fit <- predict(fit.var,newx)          
    
        var.noise <- mean( var.fit - diag(matrix(pred.t,N.fit)) ) 
    
    ### obtain the smoothed covariance functions Gw(s,t) and Gb(s,t)
        # var.noise <- mean( diag(Gw) - diag(matrix(pred,100)) )





###     Estimate eigen values and eigen functions at two levels by calling the 
###     eigen function (in R "base" package) on discretized covariance matrices.
    e1 <- eigen(Gb)
    e2 <- eigen(Gw)

###    transform the eigenvalues of the covariance matrices into eigenvalues
###    of original functions
    fpca1.value <- e1$values* tlength / N.fit
    fpca2.value <- e2$values* tlength/ N.fit 
###     Keep only non-negative eigenvalues
    fpca1.value <- ifelse(fpca1.value>=0, fpca1.value, 0)
    fpca2.value <- ifelse(fpca2.value>=0, fpca2.value, 0)
###     Calculate the percentage of variance that are explained by the components
    percent1 <- (fpca1.value)/sum(fpca1.value)
    percent2 <- (fpca2.value)/sum(fpca2.value)
    
###     Decide the number of components that are kept at level 1 and 2. The general
###     rule is to stop at the component where the cumulative percentage of variance 
###     explained is greater than 90% and the variance explained by any single component
###     after is less than 1/N. The number of components are also no less than the 
###     pre-determined minimum values min.K1 or min.K2.
    K1 <- max( which(cumsum(percent1) < 0.9 | percent1 > 1/N.fit ) + 1, min.K1 )
    K2 <- max( which(cumsum(percent2) < 0.9 | percent2 > 1/N.fit ) + 1, min.K2 )

###     estimate eigen vectors for discretized covariance matrices and
###     transform them into norm one eigenfunctions
    fpca1.vectors <- e1$vectors[, 1:K1]*sqrt(N.fit/tlength)
    fpca2.vectors <- e2$vectors[, 1:K2]*sqrt(N.fit/tlength)


###     The eigenfunctions are unique only up to a change of signs.
###     Select the signs of eigenfunctions so that the integration over the domain 
###     is non-negative
    for(i in 1:K1) {
        v2 <- fpca1.vectors[,i]
        tempsign <- sum(v2)
        fpca1.vectors[,i] <- ifelse(tempsign<0, -1,1) * v2
    }
    for(i in 1:K2) {
        v2 <- fpca2.vectors[,i]
        tempsign <- sum(v2)
        fpca2.vectors[,i] <- ifelse(tempsign<0, -1,1) * v2
    }


lam1=fpca1.value[1:K1]
lam2=fpca2.value[1:K2]
phi1=fpca1.vectors
phi2=fpca2.vectors


###
###     Estimatimate principal component scores and predict various functions
###

library(MASS) 

index <- matrix(0, M*J, N)
mu.hat <- matrix(0, M*J, N)    
y.center <- matrix(0, M*J, N)

phi1.hat <- array(0, dim=c( M*J, N, K1) )
phi2.hat <- array(0, dim=c( M*J, N, K2) )

t.fit.matrix <- matrix(rep(t.fit, each=N), nrow=N)

for(m in 1:M ) {
    for(j in 1:J) {
        row.ind <- (m-1)*J + j
        index <- apply( (t.fit.matrix - tx[row.ind, ])^2, 1, which.min )
        y.center[row.ind, ] <- y[ row.ind ,] - mu.fit[ index ] - eta.fit[ index , j] 
        phi1.hat[row.ind, , ] <- phi1[index, ]
        phi2.hat[row.ind, , ] <- phi2[index, ]
    }
}



est.si1 <- matrix(0, nrow=M, ncol=K1)
est.si2 <- matrix(0, nrow=M*J, ncol=K2)

est.si1.sd <- matrix(0, nrow=M, ncol=K1)
est.si2.sd <- matrix(0, nrow=M*J, ncol=K2)


y.hat.center <- matrix(0, M*J, N.fit) 
y.hat <- matrix(0, M*J, N.fit)
y.hat.subject <- matrix(0, M, N.fit)

y.hat.sd <- matrix(0, M*J, N.fit)
y.hat.subject.sd <- matrix(0, M, N.fit)



for(m in 1:M) {

cov.y <- matrix(0, N*J, N*J)
cov.xi.y <- matrix(0, K1 + J*K2,  N*J )

    for(j1 in 1:J) {
        ind1 <- J*(m-1) + j1
        cov.y [ 1:N + (j1-1) *N , 1:N + (j1-1) *N ] <- phi1.hat[ind1,,] %*% diag( lam1) %*%  t( phi1.hat[ind1,,] ) + phi2.hat[ind1,,] %*% diag( lam2) %*%  t( phi2.hat[ind1,,] ) + diag( rep(var.noise, N) )
        
        if(j1<J) {
            for(j2 in (j1+1):J ) {
                ind2 <- J*(m-1) + j2
                cov.y[ 1:N + (j2-1) *N , 1:N + (j1-1) *N ] <- phi1.hat[ind2,,] %*% diag( lam1) %*%  t( phi1.hat[ind1,,] )
                cov.y[ 1:N + (j1-1) *N , 1:N + (j2-1) *N ] <- t( cov.y[ 1:N + (j2-1) *N , 1:N + (j1-1) *N ] )
            }
        } 
    }

    for(j in 1:J) {
        ind1 <- J*(m-1) + j
        cov.xi.y[1:K1, 1:N + (j-1) *N] <- diag( lam1) %*% t( phi1.hat[ind1,,] )
        cov.xi.y[ K1 + (1:K2) + K2* (j-1), 1:N + N*(j-1) ] <- diag( lam2) %*% t( phi2.hat[ind1,,] )  
    }

temp.mat <- cov.xi.y %*% ginv(cov.y)
score <- temp.mat %*% as.vector( t(y.center[J*(m-1) + 1:J,]) )
var.score <- diag( c(lam1, rep(lam2, J) ) ) -  temp.mat %*% t(cov.xi.y)

est.si1[m ,] <- score[1:K1]
est.si2[ 2*(m-1) + 1:J  ,] <- t( matrix( score[ K1 + (1: (K2*J) ) ], ncol=J) )

y.hat.subject[m, ] <- mu.fit + eta.fit[,j] + phi1 %*% est.si1[m,] 
y.hat.subject.sd[m,] <-  sqrt( diag( phi1 %*% var.score[1:K1, 1:K1] %*% t( phi1 ) ) )

    for(j in 1:J) {
        row.ind <- (m-1)*J + j
        y.hat.center[ row.ind ,] <-  phi1 %*% est.si1[m,] + phi2 %*% est.si2[row.ind ,] 
        y.hat[row.ind,] <- y.hat.center[ row.ind ,]  +  mu.fit + eta.fit[,j] 
        
        temp1 <- cbind( phi1, phi2 )
        temp2 <- var.score[ c(1:K1, K1 + (j-1)* K2 + (1:K2)  ) , c(1:K1, K1 + (j-1)* K2 + (1:K2)  ) ]
        y.hat.sd[row.ind, ] <- sqrt( diag( temp1 %*% temp2 %*% t(temp1) ) )   
    }

}


###     Return the results from multilevel FPCA as a list
    return( list(K1=K1, K2=K2, t.fit=t.fit, N.fit=N.fit, lambda1=fpca1.value, lambda2=fpca2.value, phi1=fpca1.vectors, 
        phi2=fpca2.vectors, mu=mu.fit, eta=eta.fit, var.noise=var.noise, y.hat=y.hat, y.hat.sd=y.hat.sd, 
        y.hat.subject=y.hat.subject, y.hat.subject.sd=y.hat.subject.sd ) )

} 
