

########################################################################################
###         This file contains R codes to illustrate an example of
###               MFPCA (Multilevel Functional Principal Component Analysis)
###                     for sparse multilevel functional data
###                                     
###             Author:     Chongzhi Di
###             Date:       08/16/2009
###
###                                                             
###             References:
###                 (1) Di, Crainiceanu, Caffo and Punjabi (2009), Multilevel Functional Principal Component Analysis.
###                     The Annals of Applied Statistics, Vol. 3, No. 1, 458488.
###                 (2) Crainiceanu, Staicu and Di (2009), Generalized Multilevel Functional Regression. JASA, to appear.
###                 (3) Di and Crainiceanu (2009), Multilevel Functional Principal Component Analysis for Sparsely Sampled
###                     Hierarchical Functions. In preparation for submission. 
###            
########################################################################################



########################################################################################
###     Part I: Simulate a dataset of sparse multielvel functions
########################################################################################

###     Set the seed for simulation
set.seed(1)

###     K1: number of true level 1 principal components
###     K2: number of true level 2 principal components 
###     K = K1 + K2 : number of total principal components
###     The true data is generated with K1=4 and K2=4

###     N: number of grid points for each function
###     M: number of clusters of functions
###     J: number of functions per cluster

###     N.fit:  the number of grid points for the underlying true functions
###     t.fit:  the set of grid points for the underlying true functions

K1 <- 4
K2 <- 4
K <- K1 + K2
M <- 100
J <- 2

N <- 4
sigma <- 0.5


N.fit <- 50
t.fit <- seq(0, 1, length=N.fit)


### Create id and visit variables for rows of data matrix below
### The data matrix will be organized in the following order (with 100 subjects and 2 visits per subject)
###     row 1: subject 1 visit 1
###     row 2: subject 1 visit 2
###     row 3: subject 2 visit 1
###     row 4: subject 2 visit 2
###     ......

id <- rep(1:M, each=J)
visit <- rep(1:J, M)


###     generate the true eigenvalues -- lambda1 and lambda2
lambda1 <-  0.5^(0:(K1-1))
lambda2 <-  0.5^(0:(K2-1))


###   The number of grid points per function   
n.grid <- rep(N, M*J)


###     Generate the random set of grid points per function
###         by randomly selecting "N" grid points out of "N.fit" possible grid points
###         (in the example code, randomly select 3 out of 50 grid points per function)
tx <- matrix(0, M*J, N)
for(i in 1:(M*J) ) {
    tx[i,] <- sort( runif(N) )
}


###     The true mean function
mu.true <- function(x) {    return( 8*x*(1-x) )  }

###     The true level 1 eigenfunctions are four muturally orthogonal 
###     sine and cosine functions
f1 <- function(x) {
    res <- matrix(0, length(x), 4)
    for ( i in 1:2 ) {
        res[,2*i-1] <- sqrt(2)*sin(i*x*2*pi)
        res[,2*i] <- sqrt(2)*cos(i*x*2*pi)
    }
    return(res)
}

###     The true level 2 eigenfunctions are four muturally orthogonal 
###     polynomial functions

f2 <- function(x) {
    res <- matrix(0, length(x), 4)
    res[,1] <- rep(1, length(x) )*sqrt(1)
    res[,2] <- sqrt(3) * (2*x - 1)
    res[,3] <- sqrt(5) * (6*x^2 - 6 * x + 1)
    res[,4] <- sqrt(7) * (20*x^3 - 30*x^2 + 12 * x -1)
    return(res)
}


###     Generate subject and visit level principal component scores from normal distributions 
###          with variance equal to the corresponding eigenvalue.
si1 <- matrix(0, nrow=M, ncol=K1)
si2 <- matrix(0, nrow=M*J, ncol=K2)
for(k in 1:K1) {
    si1[,k] <- rnorm( M,sd=sqrt(lambda1[k]) )
}
for(k in 1:K2) {
    si2[,k] <- rnorm( M * J ,sd=sqrt(lambda2[k]) )
}

###     Generate the observed functions according to the KL expansions
###         y: observed functions
###         y.fit:          the true functions evaluated at the set of grid points "t.fit"
###         y.fit.subject:  the subject level true functions evaluated at the set of grid points "t.fit"
y <- matrix(0, M*J, N)
y.fit <- matrix(0, M*J, N.fit)
y.fit.subject <- matrix(0, M, N.fit)

for(m in 1:(M) ) {
    for(j in 1:J) {
        row.ind <- (m-1)*J + j
        temp.x <- tx[row.ind,]
        y[ row.ind ,] <- mu.true(temp.x) +  f1(temp.x) %*% si1[m,] + f2(temp.x)%*% si2[row.ind ,] + rnorm(N, mean=0, sd=sigma) 
        y.fit.subject[m, ] <- mu.true(t.fit) +  f1(t.fit) %*% si1[m,]
        y.fit[ row.ind ,] <- mu.true(t.fit) +  f1(t.fit) %*% si1[m,] + f2(t.fit)%*% si2[row.ind ,] 
    }
}


### plot observed (dots) vs true (lines) functions for the first three subjects
par(mfrow = c(3,2), mar=c(2,2,3,2))
for(m in 1:3) 
    for(j in 1:2) {
        row.ind <- (m-1)*J + j
        temp.x <- tx[row.ind,]
        plot(t.fit, y.fit[row.ind,],type="l", ylim=c(-4.5,8), main=paste("subject", m, "visit", j)  )
        points(temp.x, y[row.ind,])
    }







########################################################################################
###
###     Part II: Call "MFPCA.sparse" function to conduct the MFPCA analysis
###
########################################################################################

###     Set the directory to the one that includes the file "MFPCA-sparse.R"
setwd("C:/Documents and Settings/Cdi/My Documents/_Sleep/sleepsoftware/FPCA-sparse/Version 2 - shared/")
source("MFPCA-sparse-new.R")

###     Set the knots for univariate and bivariate smoothing
###     Due to restrictions of the function "spm" from package "SemiPar",
###         these must be specified outside to work properly
library(SemiPar)
N.smooth.single=20
N.smooth=15
t.knot <- seq(0, 4, length=N.smooth.single)
myknots.t <- data.frame(x1t=rep(seq(0,4,length=N.smooth),each=N.smooth), x2t=rep(seq(0,4,length=N.smooth),N.smooth)) 
myknots.b <- data.frame(x1b=rep(seq(0,4,length=N.smooth),each=N.smooth), x2b=rep(seq(0,4,length=N.smooth),N.smooth))

###     Call the "MFPCA.sparse" function
fit.sparse <- MFPCA.sparse(y, tx, M=M, J=2, N=N, twoway=TRUE, tstart=0, tlength=1)


###      plot estimated mean functions: overall, visit 1, and visit 2
time.grid <- fit.sparse$t.fit
par(mfrow=c(1,1))
plot(time.grid, fit.sparse$mu, ylim=c(-5,5), type="l",main="mean functions", lwd=2, 
    xlab="x", ylab="")
lines(time.grid, fit.sparse$mu + fit.sparse$eta[,1], lty=2)
lines(time.grid, fit.sparse$mu + fit.sparse$eta[,2], lty=3)
legend(0.3, -0.5, legend=c("overall", " visit 1", " visit 2", "true mean"), lty=c(1,2,3,1),
        col=c("black", "black", "black", "red"), lwd=c(2,1,1,1), bty="n")
lines(time.grid, mu.true(time.grid), col="red")



###     Estimated eigenvalues: level 1 and 2
sum.lambda1 <- sum(fit.sparse$lambda1)
percent.var1 <- fit.sparse$lambda1 / sum.lambda1
sum.lambda2 <- sum(fit.sparse$lambda2)
percent.var2 <- fit.sparse$lambda2 /sum.lambda2
res.lam1 <- round( rbind(1*fit.sparse$lambda1, 100*fit.sparse$lambda1/sum.lambda1,
        100*cumsum(fit.sparse$lambda1)/sum.lambda1),digits=3 )[, 1:fit.sparse$K1]
rownames(res.lam1) <- c("eigenvalues", "percent", "cum.percent")
res.lam2 <- round( rbind(1*fit.sparse$lambda2, 100*fit.sparse$lambda2/sum.lambda2,
        100*cumsum(fit.sparse$lambda2)/sum.lambda2),digits=3 )[, 1:fit.sparse$K2]
rownames(res.lam2) <- c("eigenvalues", "percent", "cum.percent")

###      percentage of variation explained by level 1 
rho <- c(sum.lambda1,sum.lambda2)/(sum.lambda1 + sum.lambda2)

res.lam1
res.lam2
rho


###     Estimated level 1 eigenvalues: plot the first 4 components
###         red lines: true eigenfunctions
###         black lines: estimated eigenfunctions

par(mfrow=c(2,2),mar=c(3,3,3,2))
for(i in 1:4) {
    v1 <- fit.sparse$phi1[,i]
    v2 <- f1(time.grid)[,i]
    sign.v1 <- ifelse( sum((v1-v2)^2)<sum((v1+v2)^2), 1, -1 )
    v1 <- sign.v1 * v1
    plot(time.grid,v1, type="l", ylab="",  xlab="x", ylim=c(-2,2), lwd=2,
        main=paste("level 1 comp ", i,", ",100*round(percent.var1[i],3), "%", sep=""))
    lines(time.grid, v2, col="red")
}
par(mfrow=c(1,1))


###     Estimated level 2 eigenfunctions: plot the first 4 components
###         red lines: true eigenfunctions
###         black lines: estimated eigenfunctions

par(mfrow=c(2,2),mar=c(3,3,3,2))
for(i in 1:4) {
    v1 <- fit.sparse$phi2[,i]
    v2 <- f2(time.grid)[,i]
    sign.v1 <- ifelse( sum((v1-v2)^2)<sum((v1+v2)^2), 1, -1 )
    v1 <- sign.v1 * v1
    plot(time.grid,v1, type="l", ylab="",  xlab="x", ylim=c(-2,2), lwd=2,
        main=paste("level 2 comp ", i,", ",100*round(percent.var1[i],3), "%", sep=""))
    lines(time.grid, v2, col="red")
}
par(mfrow=c(1,1))
  


###     Plot and prediction of subject specific and subject/visit specific curves
###         We plotted 3 subjects, 2 visits per subject
###         Each column corresponds to a subject
###             row 1: true functions (red lines), observed functions (blue dots) and predicted functions (thick dashed lines)
###                    with confidence bands (thin dashed lines) for visit 1 (subject/visit specific)
###             row 2: true functions (red lines), observed functions (blue dots) and predicted functions (thick dashed lines)
###                    with confidence bands (thin dashed lines) for visit 2 (subject/visit specific)
###             row 3: true functions (red lines) and predicted functions (thick dashed lines)
###                    with confidence bands (thin dashed lines) at subject specific level (unobservable)


attach(fit.sparse, pos=3)

par(mfcol = c(3,3), mar=c(2,2,3,2))
for(m in 1:3) {

for(j in 1:2) {
        row.ind <- (m-1)*J + j
        temp.x <- tx[row.ind,]
        plot(t.fit, y.fit[row.ind,],type="l", ylim=c(-5,7.5), main=paste("sub ", m, " visit ", j, ": N=", N, sep=""), col="red", lwd=2 )
        points(temp.x, y[row.ind, ], col="blue", pch=19, cex=1.5)
        lines(t.fit, y.hat[row.ind,], lty=2, lwd=2) 
        lines(t.fit, y.hat[row.ind,] - 1.96*y.hat.sd[row.ind,], lty=2)
        lines(t.fit, y.hat[row.ind,] + 1.96*y.hat.sd[row.ind,], lty=2)
    }
        plot(t.fit, y.fit.subject[m,],type="l", ylim=c(-5,7.5), main=paste("sub ", m, ": N=", N, sep="") , col="red", lwd=2 )
        lines(t.fit, y.hat.subject[m,], lty=2, lwd=2) 
        lines(t.fit, y.hat.subject[m,] - 1.96*y.hat.subject.sd[m,], lty=2)
        lines(t.fit, y.hat.subject[m,] + 1.96*y.hat.subject.sd[m,], lty=2)
        
}

detach(fit.sparse, pos=3)
